//
// Created by delanoy on 2/9/18.
//

#include "CSG_graph.hpp"

#include <igl/readOFF.h>
#include <igl/per_face_normals.h>


void setSymmetricPosition(std::shared_ptr<Primitive> p, std::string currentOFF,
                          std::default_random_engine & generator, int sym_axis)
{
    //load previous iteration off object
    Eigen::MatrixXd V, N;
    Eigen::MatrixXi F, F_row;
    igl::readOFF(currentOFF, V, F);
//randomly choose  a face
    std::uniform_int_distribution<int> distr_face(0, F.rows()-1);
    int f_id = distr_face(generator);
    Eigen::RowVector3d center = (V.row(F(f_id, 0)) + V.row(F(f_id, 1)) + V.row(F(f_id, 2)))/3;
    //compute face normals
    F_row = F.row(f_id);
    igl::per_face_normals(V, F_row, N);
    //set the distance of the new primitive
    std::uniform_real_distribution<float> distr_dist(0, 0.7*p->distFromCenter(N.row(0)));
    Eigen::Vector3d pos = center + distr_dist(generator) * N.row(0);
    //if too close from the center, put in the center
    if (center(sym_axis) < 0.5)
        pos(sym_axis) = 0;
    //check that the 2 primitives have a sufficient shift
    Eigen::Vector3d dir_plane(0,0,0);
    dir_plane(sym_axis) = 1;
    float dist_plane = std::abs(pos(sym_axis))-p->distFromCenter(dir_plane);
    if(dist_plane < 0.5 && dist_plane > 0)
        pos(sym_axis)=dist_plane+1;

    p->setPos(pos);
}



std::shared_ptr<CSG_node> CSG_node::read_graph(std::string filename)
{
    std::stack<CSG_operator> op_stack;
    std::stack<std::shared_ptr<CSG_node> > node_stack;
    std::shared_ptr<CSG_node> left_node;
    std::shared_ptr<CSG_node> right_node;

    std::ifstream script;
    script.open(filename);
    //read and push all operators
    std::string line;
    while (std::getline(script, line))
    {
        // std::cout<<line<<std::endl;
        if (line.find("union") != std::string::npos)
        {
            // std::cout<<"finding union\n";
            op_stack.push(CSG_operator::UNION);
        } else if (line.find("diff") != std::string::npos)
        {
            // std::cout<<"finding diff\n";
            op_stack.push(CSG_operator::DIFF);
        } else if (line.find("}") != std::string::npos)
        {
            // std::cout<<"finding }\n";
            std::shared_ptr<CSG_node> right_node = node_stack.top();
            // std::cout<<"-----right :"<<*right_node<<std::endl;
            node_stack.pop();
            std::shared_ptr<CSG_node> left_node = node_stack.top();
            // std::cout<<"-----left :"<<*left_node<<std::endl;
            node_stack.pop();
            std::shared_ptr<CSG_node> n(new CSG_graph(op_stack.top(), left_node, right_node));
            // std::cout<<"-----new :"<<*n<<std::endl;
            op_stack.pop();
            node_stack.push(n);
        } else
        {
            // std::cout<<"finding prim\n";
            std::shared_ptr<Primitive> p;
            //must be on a primitive, try to extract values
#define CHARS "%*[^0-9-.]"
            float pos[3], size[3], rot[3];
            int nb = sscanf(line.c_str(), CHARS "%f" CHARS "%f" CHARS "%f" CHARS "%f" CHARS "%f" CHARS "%f" CHARS "%f" CHARS "%f" CHARS "%f", pos, pos+1, pos+2, size, size+1, size+2, rot, rot+1, rot+2);
            if (line.find("sphere") != std::string::npos)
            {
                p = std::shared_ptr<Primitive>(new Sphere());
                p->setSize(size[0], size[1], size[2]);
            } else if (line.find("cube") != std::string::npos)
            {
                p = std::shared_ptr<Primitive>(new Cube());
                p->setSize(size[0]/2, size[1]/2, size[2]/2);
            } else if (line.find("cylinder") != std::string::npos)
            {
                p = std::shared_ptr<Primitive>(new Cylinder());
                p->setSize(size[0], size[1], size[2]);
                std::dynamic_pointer_cast<Cylinder>(p)->setRotation(rot[0], rot[1], rot[2]);
            } else
            {
                std::cerr<<"Unknown primitive \n";
            }
            p->setPos(pos[0], pos[1], pos[2]);
            std::shared_ptr<CSG_node> n(new CSG_leave(p));
            // std::cout<<"-----new :"<<*n<<std::endl;
            node_stack.push(n);
        }
    }
    script.close();
    return node_stack.top();
}

std::shared_ptr<CSG_node>
CSG_node::generate_CSG_tree(std::default_random_engine &generator,
                            int nb_prim, double max_size, std::string tmp_file)
{
    std::cout<<"------------ "<<nb_prim<<" primitives\n";
    std::shared_ptr<CSG_node> current_graph, last_graph;
    //TODO choose alignment axis
    int sym_axis = 0;
    int i = 0;
    while(i < nb_prim)
    {
        //generate primitive and its copy (symmetric)
        std::uniform_int_distribution<int> prim_type_gen(1,2);
        std::shared_ptr<Primitive> s, s2;
        switch (prim_type_gen(generator)) //no sphere
        {
            case 1:
                s = Cube::getRandom(generator, max_size);
                s2 = std::shared_ptr<Primitive>(new Cube(*(Cube*)s.get()));
                break;
            case 2:
                s = Cylinder::getRandom(generator, max_size);
                s2 = std::shared_ptr<Primitive>(new Cylinder(*(Cylinder*)s.get()));
                break;
        }

        //add it to the graph
        if (i == 0)
        {
            //start the graph
            std::shared_ptr<CSG_leave> l(new CSG_leave(s));
            current_graph = l;
        }
        else
        {
            std::shared_ptr<CSG_leave> l(new CSG_leave(s));
            //find pos
            setSymmetricPosition(s, tmp_file+".off", generator,0);
            Eigen::Vector3d pos = s->getPos();
            pos[sym_axis] *= -1;
            s2->setPos(pos);
            //add the 2 primitives
            //TODO if dist to plane of sym > threshold, add one centered primitive
            std::uniform_real_distribution<float> rand_op(0,1.0);
            CSG_node::CSG_operator op;
            if(rand_op(generator) > 0.7) //more union that dif
                op = CSG_node::CSG_operator::DIFF;
            else
                op = CSG_node::CSG_operator::UNION;
            current_graph = std::shared_ptr<CSG_graph>
                    (new CSG_graph(op, current_graph, l));
            //put symmetric primitive
            std::shared_ptr<CSG_leave> l2(new CSG_leave(s2));
            current_graph = std::shared_ptr<CSG_graph>
                    (new CSG_graph(op, current_graph, l2));

        }

        //rendering graph into OFF file
        std::ofstream file_tmp(tmp_file+".escad");
        file_tmp<<*current_graph;
        file_tmp.close();
        std::cout<<"generate off"<<std::endl;
        std::string cmd = OPENSCAD_PATH" -o "+tmp_file+".off -D '$fa=24' -D '$fs=0.5' "+tmp_file+".escad";
        system(cmd.c_str());

        if (i > 0)
        {
            //verify that the object is not empty
            Eigen::MatrixXd V;
            Eigen::MatrixXi F;
            igl::readOFF(tmp_file+".off", V, F);
            if (V.rows() == 0)
            {
                std::cerr<<"EMPTY OBJECT, get back of one step!"<<std::endl;
                current_graph = last_graph;
                i--;
            }
        }
        i++;
        last_graph=current_graph;

    }
    return current_graph;
}
