//
// Created by delanoy on 2/9/18.
//

#ifndef ESCAD_PROJECT_PRIMITIVE_HPP
#define ESCAD_PROJECT_PRIMITIVE_HPP



#include <iostream>
#include <Eigen/Core>
#include <vector>
#include <memory>

class Primitive
{
protected:
    Eigen::RowVector3d _pos;
    Eigen::RowVector3d _size;

    void printTranslation(std::ostream& str) const
    {
        str << "translate (["<<_pos(0)<<","<<_pos(1)<<","<<_pos(2)<<"]) ";
    }
    void printVector(std::ostream& str, Eigen::RowVector3d v) const
    {
        str << "["<<v(0)<<","<<v(1)<<","<<v(2)<<"]";
    }
public:
    Primitive() :
            _pos(0,0,0), _size(1,1,1) {}
    Primitive(double pos_x, double pos_y, double pos_z) :
            _pos(pos_x, pos_y, pos_z), _size(1,1,1) {}
    Primitive(const Eigen::Vector3d &pos) :
            _pos(pos), _size(1,1,1) {}

    void setPos(double x, double y, double z)
    {
        _pos = Eigen::RowVector3d(x, y, z);
    }
    void setPos(const Eigen::Vector3d &pos)
    {
        _pos = Eigen::RowVector3d(pos);
    }
    Eigen::Vector3d getPos() const
    {
        return _pos;
    }
    void setSize(double x, double y, double z)
    {
        _size = Eigen::RowVector3d(x, y, z);
    }



    //need to be implemented in children
    virtual double distFromCenter(Eigen::Vector3d dir) const =0; //return the distance from the center to the surface following direction dir

    static std::shared_ptr<Primitive> getRandom(std::default_random_engine &generator,
                                                double max_size)
    {
        return std::shared_ptr<Primitive>(NULL);
    }

    friend std::ostream& operator<<(std::ostream& str, const Primitive &p)
    {
        return p.printScript(str);
    }

private:
    virtual std::ostream& printScript(std::ostream& str) const = 0;

};


class Sphere : public Primitive
{
private:
    //double _radius;
public:
    Sphere() : Primitive() {}
    Sphere(double pos_x, double pos_y, double pos_z) : Primitive(pos_x, pos_y, pos_z){}
    Sphere(const Eigen::Vector3d &pos) : Primitive(pos) {}

    double distFromCenter(Eigen::Vector3d dir) const
    {
        double eq = ((dir(0)*dir(0))/(_size(0)*_size(0))) + ((dir(1)*dir(1))/(_size(1)*_size(1)))+((dir(2)*dir(2))/(_size(2)*_size(2)));
        return sqrt(1.0/eq);
    }

    static std::shared_ptr<Primitive> getRandom(std::default_random_engine & generator,
                                                double max_size)
    {
        std::uniform_real_distribution<float> distr_size(0.2*max_size, max_size);
        std::shared_ptr<Primitive> s(new Sphere());
        s->setSize(distr_size(generator), distr_size(generator), distr_size(generator));
        return s;
    }

    friend std::ostream& operator<<(std::ostream& str, const Sphere &p)
    {
        return p.printScript(str);
    }
private:
    std::ostream& printScript(std::ostream& str) const
    {
        printTranslation(str);
        str << "scale (";
        printVector(str,_size);
        return str << ") sphere(1);\n";
    }
};



class Cube : public Primitive
{
private:
    //double _radius;
public:
    Cube() : Primitive() {}
    Cube(double pos_x, double pos_y, double pos_z) : Primitive(pos_x, pos_y, pos_z){}
    Cube(const Eigen::Vector3d &pos) : Primitive(pos) {}

    double distFromCenter(Eigen::Vector3d dir) const
    {

        return std::min(_size(0)/std::abs(dir[0]), std::min(_size(1)/std::abs(dir[1]), _size(2)/std::abs(dir[2])));
    }

    static std::shared_ptr<Primitive> getRandom(std::default_random_engine & generator,
                                                double max_size)
    {
        std::uniform_real_distribution<float> distr_size(0.2*max_size, max_size);
        std::shared_ptr<Primitive> s(new Cube());
        s->setSize(distr_size(generator), distr_size(generator), distr_size(generator));
        return s;
    }

    friend std::ostream& operator<<(std::ostream& str, const Cube &p)
    {
        return p.printScript(str);
    }
private:
    std::ostream& printScript(std::ostream& str) const
    {
        printTranslation(str);
        str << "cube (";
        printVector(str,_size*2);
        return str<<", center=true) ;\n";
    }
};



class Cylinder : public Primitive
{
private:
    //double _radius;
    Eigen::RowVector3d _rotate;
public:
    Cylinder() :
            Primitive(), _rotate(0,0,0) {}
    Cylinder(double pos_x, double pos_y, double pos_z) :
            Primitive(pos_x, pos_y, pos_z), _rotate(0,0,0){}
    Cylinder(const Eigen::Vector3d &pos) :
            Primitive(pos), _rotate(0,0,0){}

    double distFromCenter(Eigen::Vector3d dir) const
    {
        //find revolution axis
        int axis[3]={2,1,0};
        if(_rotate(0) > 0)
            axis[0]=1;
        else if (_rotate(1) > 0)
            axis[0]=0;
        axis[1] = (axis[0] + 1)%3;
        axis[2] = (axis[0] + 2)%3;
        double dist_flat = _size(axis[0])/std::abs(dir[axis[0]]);
        double dist_round = ((dir(axis[1])*dir(axis[1]))/(_size(axis[1])*_size(axis[1]))+
                             (dir(axis[2])*dir(axis[2]))/(_size(axis[2])*_size(axis[2])));
        return std::min(dist_flat, sqrt(1.0/dist_round));
    }

    static std::shared_ptr<Primitive> getRandom(std::default_random_engine & generator,
                                                double max_size)
    {
        std::uniform_real_distribution<float> distr_size(0.2*max_size, max_size);
        std::shared_ptr<Cylinder> s(new Cylinder());
        s->setSize(distr_size(generator), distr_size(generator), distr_size(generator));
        std::uniform_int_distribution<int> distr_axis(0,2);
        s->_rotate(distr_axis(generator)) = 90;
        return s;
    }
    void setRotation(double x, double y, double z)
    {
        _rotate = Eigen::RowVector3d(x, y, z);
    }
    void setRotation(const Eigen::Vector3d &pos)
    {
        _rotate = Eigen::RowVector3d(pos);
    }
    friend std::ostream& operator<<(std::ostream& str, const Cylinder &p)
    {
        return p.printScript(str);
    }
private:
    std::ostream& printScript(std::ostream& str) const
    {
        printTranslation(str);
        str << "scale (";
        printVector(str,_size);
        str << ") rotate (";
        printVector(str,_rotate);
        str << ") cylinder (h=2, r=1, center=true) ;\n";
        return str;
    }
};

#endif //ESCAD_PROJECT_PRIMITIVE_HPP
